;===============================================
; FiveSquare.asm
;
;
; LED 5x5 LED Matrix Application
;
; Microchip Assembly program
; June 1, 2003
; June 6, 2008: Redone as FiveSquare
; David D. Kaufman
;===============================================

; -- Future features --
; Port 'tile' code from longfield

#if 0

	Physical Organization...
	
		_04_  _03_  _02_  _01_  _00_
		_09_  _08_  _07_  _06_  _05_
		_0e_  _0d_  _0c_  _0b_  _0a_
		_13_  _12_  _11_  _10_  _0f_
		_18_  _17_  _16_  _15_  _14_


	Memory...
	1. Packed Storage
	   Four (4) bytes are used to store individual LED 
	   values. The byte at the lower address stores
	   values _0_ to _7_. The byte at the higher
	   address stores _8_ to _f_.
	   
		Bit representation:
		07 06 05 04 03 02 01 00
		0f 0e 0d 0c 0b 0a 09 08
		17 16 15 14 13 12 11 10
		xx xx xx xx xx xx xx 18
		 
		Diagram showing row packing
		1 1 1 0 0 0 0 0
		3 2 2 2 2 2 1 1
		4 4 4 4 3 3 3 3
		. . . . . . . 4

	2. Expanded Storage ("Ready for Display")
	   Five (5) bytes are used to store the values, one
	   per row. The five bits in each byte
	   store the LED values. This data is written directly
	   to the port.

		xx 04 xx xx 03 02 01 00
		xx 09 xx xx 08 07 06 05
		xx 0e xx xx 0d 0c 0b 0a
		xx 13 xx xx 12 11 10 0f
		xx 18 xx xx 17 16 15 14
	
	Implementation...
	   For each call to an animation routine, the
	   routine will update the packed storage memory.
	   The application will expand the packed storage
	   and update the expanded storage. This will be
	   used to drive the display for a fixed period
	   of time. 

	   PORT A: Columns
	   PORT B: Rows
	   
	Animation Models...
	1. Frame-by-Frame
	   Each step of the display is stored as two
	   bytes.  
	   
	2. Pixel List ("Shift Register Style")
	   The data for the animation is stored as a list
	   of LED ids (_0_ to _f_), two per byte. For each
	   step of the animation, the current LED id is
	   added to the list.
	
	
#endif

		;---------------------------------------------
		;	Specify Processor and include file
		;---------------------------------------------
		processor 16f648a

#include "P16F648a.INC"

		

		;---------------------------------------------
		;	Adjust warnings
		;---------------------------------------------
		errorlevel -302
		
		;---------------------------------------------
		;	Decimal
		;---------------------------------------------
		radix	dec
		
		
		;---------------------------------------------
		;	Chip configuration
		;---------------------------------------------
		;	Oscillator: Internal oscillator with addition i/o line
		;	Watchdog Timer: Off
		;	Power-up timer: On
		;	Brown Out Detect: On
		;	Master Clear Enable: Disabled
		;	Low-voltage programming: Disabled
		;	Data EE Read Protect: Disabled
		;	Code Protect: Off
		;---------------------------------------------
		__config _INTRC_OSC_NOCLKOUT & _WDT_OFF & _PWRTE_ON & _BODEN_ON & _MCLRE_OFF & _LVP_OFF

		;---------------------------------------------
		;	Macros
		;---------------------------------------------
		
		; Macro:		IdxJump
		; Description:	This macro puts the high byte of the label into PCLATH
		; 				and uses the file register as a value for a computed jump
		; Params:		CallLabel - The label of the routine
		;				IndexReg - The file register which stores the offset value
		;
IdxJump		macro CallLabel, IndexReg
			movlw		HIGH CallLabel
			movwf		PCLATH
			movf		IndexReg,W
			addwf		PCL,F
			endm

									; Microchip AN Table Read 00556e.pdf
IdxLongJump 	macro CallLabel, IndexReg
			; CALL.LABEL MUST BE AT THIS LINE
			movf		IndexReg,w
			addlw		LOW (CallLabel+9)	; get low 8 bits of address
			movwf		TABLE_REG
			movlw		HIGH (CallLabel+9)	; get high 5 bits of address
			btfsc		STATUS,C	; page crossed?
			addlw		1			; yes then increment high address
			movwf		PCLATH		; load high address in latch
			movf		TABLE_REG,w	; load computed offset in w reg
			movwf		PCL
			; DATA STARTS AT THIS LINE
			endm
			
		; Macro:		PixelListAnim
		; Description:	This macro creates a pixel list animation routine.
		;				No additional statements are required or allowed. 
		; Params:		RoutineName - The label of the animation routine
		;				RepeatCnt   - The number of times to repeat the animation
		;				SpeedValue  - The speed value (smaller is faster)
		;				RunLength   - The number of pixels to turn on before turning
		;						      pixels off
		;				DataName    - The name of the routine that provides the pixel list data
		;
PixelListAnim macro RoutineName, RepeatCnt, SpeedValue, RunLength, DataName, PixelListStyle
RoutineName	IdxLongJump RoutineName, CURCMD
			retlw		RepeatCnt	; Repeat
			retlw		SpeedValue	; Duration (speed)
			movlw		RunLength	; Store runlength here
			movwf		RUNLENGTH	; ..there isn't a cmd for it
			if PixelListStyle==PixelListWrapStyle
			bsf			PIXELFLAGS, FLAG_WRAP_STYLE
			endif
			btfss		SEQFLAGS, FLAG_GET_DATA
			goto		LoadPixelListData
            movlw		HIGH (DataName)	; get high bits of address
            movwf		PCLATH
			call		DataName
			movwf		TEMP
			movlw		HIGH ($) ; get high bits of current address
			movwf		PCLATH
			movf		TEMP,W
			return
			endm

		; Macro:		FrameByFrameAnim
		; Description:	This macro creates a frame-by-frame animation routine.
		;				No additional statements are required or allowed. 
		; Params:		RoutineName - The label of the animation routine
		;				RepeatCnt   - The number of times to repeat the animation
		;				SpeedValue  - The speed value (smaller is faster)
		;				DataName    - The name of the routine that provides the pixel list data
		;
FrameByFrameAnim macro RoutineName, RepeatCnt, SpeedValue, DataName
RoutineName	IdxLongJump RoutineName, CURCMD
			retlw		RepeatCnt	; Repeat
			retlw		SpeedValue	; Duration (speed)
			btfss		SEQFLAGS, FLAG_GET_DATA
			goto		LoadFrameData
            movlw		HIGH (DataName)	; get high bits of address
            movwf		PCLATH
			call		DataName
			movwf		TEMP
			movlw		HIGH ($) ; get high bits of current address
			movwf		PCLATH
			movf		TEMP,W
			return
			endm
			 
		; Macro:		PixelListAnim
		; Description:	This macro creates a pixel list animation routine.
		;				No additional statements are required or allowed. 
		; Params:		RoutineName - The label of the animation routine
		;				RepeatCnt   - The number of times to repeat the animation
		;				SpeedValue  - The speed value (smaller is faster)
		;				Orientation - Horizontal or Vertical
		;				DataName    - The name of the routine that provides the pixel list data
		;				StripAnimStyle - Either Single or Wrap (use PixelListWrapStyle or 
		;
StripAnim macro RoutineName, RepeatCnt, SpeedValue, Direction, DataName, StripAnimStyle
RoutineName	IdxLongJump RoutineName, CURCMD
			retlw		RepeatCnt	; Repeat
			retlw		SpeedValue	; Duration (speed)
			if (((StripAnimStyle) & StripWrapStyle) != 0)
			bsf			PIXELFLAGS, FLAG_WRAP_STYLE
			else
			bcf			PIXELFLAGS, FLAG_WRAP_STYLE
			endif
			if (((StripAnimStyle) & StripHorizontal) != 0)
			bsf			PIXELFLAGS, FLAG_HORIZONTAL
			else
			bcf			PIXELFLAGS, FLAG_HORIZONTAL
			endif
			btfss		SEQFLAGS, FLAG_GET_DATA
			goto		LoadStripData
            movlw		HIGH (DataName)	; get high bits of address
            movwf		PCLATH
			call		DataName
			movwf		TEMP
			movlw		HIGH ($) ; get high bits of current address
			movwf		PCLATH
			movf		TEMP,W
			return
			endm

		;---------------------------------------------
		;	Normal or Test Build
		;	2008.06.02: Added
		;---------------------------------------------
;#define LED_TEST_BUILD

		;---------------------------------------------
		;	Configuration
		;---------------------------------------------

#define		ROW_VALUE 0
#define		COL_VALUE 1

#if 1	;-------------------------------------------------

#define		MULTIPLEX_DIM 	ROW_VALUE
#define		PIXELDATA_DIM 	COL_VALUE

#define		MULTIPLEX_PORT	PORTA
#define		PIXELDATA_PORT	PORTB

#define		MULTIPLEX_ACTIVE_HIGH	0
#define		PIXELDATA_ACTIVE_HIGH	1

#define		PIXELDATA_NEEDS_ALIGN_TO_PORT	0
#define		MULTIPLEX_NEEDS_ALIGN_TO_PORT	1

#else	;-------------------------------------------------

#define		MULTIPLEX_DIM 	COL_VALUE
#define		PIXELDATA_DIM 	ROW_VALUE

#define		MULTIPLEX_PORT	PORTB
#define		PIXELDATA_PORT	PORTA

#define		MULTIPLEX_ACTIVE_HIGH	1
#define		PIXELDATA_ACTIVE_HIGH	0

#define		PIXELDATA_NEEDS_ALIGN_TO_PORT	1
#define		MULTIPLEX_NEEDS_ALIGN_TO_PORT	0

#endif 	;-------------------------------------------------



#define		MIRROR_HORIZONTAL	1



#define		FRAME_SIZE			5

;#define		ROW_PORT			PORTB
;#define		COL_PORT			PORTA
;#define		ROW_ACTIVE_HIGH		1
;#define		COL_ACTIVE_HIGH		0

		;---------------------------------------------
		;	Constants
		;---------------------------------------------
#define		MULTIPLEX_DELAY		20

#if MULTIPLEX_ACTIVE_HIGH
#if MULTIPLEX_NEEDS_ALIGN_TO_PORT
#define		MULTIPLEX_ZERO			b'01000000'
#define		MULTIPLEX_ONE			b'00001000'
#define		MULTIPLEX_TWO			b'00000100'
#define		MULTIPLEX_THREE			b'00000010'
#define		MULTIPLEX_FOUR			b'00000001'
#else
#define		MULTIPLEX_ZERO			b'00010000'
#define		MULTIPLEX_ONE			b'00001000'
#define		MULTIPLEX_TWO			b'00000100'
#define		MULTIPLEX_THREE			b'00000010'
#define		MULTIPLEX_FOUR			b'00000001'
#endif
#else
#if MULTIPLEX_NEEDS_ALIGN_TO_PORT
#define		MULTIPLEX_ZERO			b'00001111'
#define		MULTIPLEX_ONE			b'01000111'
#define		MULTIPLEX_TWO			b'01001011'
#define		MULTIPLEX_THREE			b'01001101'
#define		MULTIPLEX_FOUR			b'01001110'
#else
#define		MULTIPLEX_ZERO			b'00001111'
#define		MULTIPLEX_ONE			b'00010111'
#define		MULTIPLEX_TWO			b'00011011'
#define		MULTIPLEX_THREE			b'00011101'
#define		MULTIPLEX_FOUR			b'00011110'
#endif
#endif

#define		FLAG_LIST_DONE		0	; SEQFLAGS bits
#define		FLAG_DATA_DONE		1
#define		FLAG_GET_DATA		2
#define		FLAG_DATA_EXPANDED	3

#define		FLAG_PIXEL_DONE		0	; PIXELFLAGS bits
#define		FLAG_PIXELS_OFF		1
#define		FLAG_WRAP_STYLE		2
#define		FLAG_HORIZONTAL		3
#define		FLAG_FIRST_PASS		4

#define		REPEAT_CMD			0
#define		DURATION_CMD		1
#define		DATA_CMD			2

#define		PixelListSinglePass	0
#define		PixelListWrapStyle	1
#define		StripSinglePass		0
#define		StripWrapStyle		1
#define		StripVertical		0
#define		StripHorizontal		2

		;---------------------------------------------
		;	Memory Variables
		;---------------------------------------------
			cblock	0x20
			COUNTER						; Delay counter variables
			COUNTERB
			
			PACKED:4					; Packed storage.....data is packed in all bits
			EXPANDED:5					; Expanded storage...data is 5 bits per byte; represents what should be on LEDs
			DISPLAY:5					; Display storage....data is organized and adjusted for putting out on ports; e.g., row or column multiplexing, align to port, active hi/lo
			ROTATED:5					; Rotate buffer......temporary buffer used during rotations
			WORK:5						; Work buffer........temporary buffer used as input to rotations and mirroring

			REPEAT						; Repeat countdown
			FRAMECOUNT					; Countdown to show current frame
			FRAMEBCOUNT					; ..
			
			SEQFLAGS					; Set to TRUE when ...
			SEQREPEAT					; Sequence repeat count
			SEQDURATION
			
			CURSEQ						; Current animation
			CURCMD						; Current command
			CURFRAME					; Index 0 to n
			DATAINDEX					; Index into data (usually generated from CURFRAME)
			DATATEMP					; Temporary cache for data index
			RUNLENGTH					; Length of pixels to shift 'on' for pixel-list animations
			PIXEL						; Pixel value
			BITMASK						; Mask for one pixel
			PIXELOFF					; Index into data for pixels to turn off
			PIXELFLAGS					; Flags for pixel list
			TEMP
			TEMP2
			ROW_INDEX					; For Strip animations
			TMP_INDEX
			ROW_COUNT
			TABLE_REG
			
			endc

		;---------------------------------------------
		;	Vectors
		;---------------------------------------------
			org 0x00
			goto Start					; Power-on Vector


		;---------------------------------------------
		;	Main Program
		;---------------------------------------------
			org	0x05
Start		bsf		STATUS, RP0
			movlw	0x00				; All pins on PortA are output
			movwf	TRISA
			;movlw	0x7F				; Turn-on weak pull-ups on Port B
			;movwf	OPTION_REG 			; to keep them from floating
			clrf	TRISB				; All pins on PortA are output
			bcf		STATUS, RP0
			
			movlw	0x07				; Change Analog pins to Digital
			movwf	CMCON
		
MainLoop
			goto	RestartList
			
		;	Design:
		;
		;	Animation List Loop
		;	|
		;	|..Get repeat count
		;	|..List done? Go to list restart
		;	|..Store repeat count
		;	|..Get duration count and store
		;	|	
		;	|	Repeat Animation Loop  
		;	|	|
		;	|	|..Reset current frame
		;	|	|
		;	|	|	Animation Frame Loop
		;	|	|	|
		;	|	|	|..Call AnimationDispatch
		;	|	|	|..Loop done? Yes: Go to loop end
		;	|	|	|..Load duration, Call DisplayFrame
		;	|	|	|..Increment frame index
		;	|	|	|
		;	|	|	End
		;	|	|
		;	|	|..Decrement repeat count
		;	|	|..Loop done? Go to loop end
		;	|	|
		;	|	End
		;	|
		;	|..Increment animation count
		;	|
		;	|
		;	End
		;
		
RestartList
			clrf	CURSEQ

ListLoop	
			clrf	SEQFLAGS					; Clear flags to test for list done
			clrf	PIXELFLAGS					; Clear flags used by pixel list animations
			movlw	REPEAT_CMD					; Get repeat value
			movwf	CURCMD
			call	AnimationDispatch
			btfsc	SEQFLAGS,FLAG_LIST_DONE		; List done?
			goto	RestartList					; Restart list
			movwf	SEQREPEAT					; Store repeat data
			
			movlw	DURATION_CMD				; Get duration data
			movwf	CURCMD
			call	AnimationDispatch
			movwf	SEQDURATION
			
			bsf		PIXELFLAGS,FLAG_FIRST_PASS	; Indicate first pass of current animation

			clrf	PACKED						; Clear the frame data
			clrf	PACKED+1
			clrf	PACKED+2
			clrf	PACKED+3
			
RepeatLoop
			clrf	SEQFLAGS					; Clear flags to test for data done
			clrf	CURFRAME					; Reset frame index
			movlw	DATA_CMD					; Set-up data command
			movwf	CURCMD
			
FrameLoop
			call	AnimationDispatch			; Update packed storage
			btfsc	SEQFLAGS,FLAG_DATA_DONE		; Was sequence done?
			goto	FrameLoopEnd				; Go to end of loop
			btfss	SEQFLAGS,FLAG_DATA_EXPANDED	; Don't call unpack routine if data already expanded
			call	ExpandPacked				; Move data to expanded storage
			call	PrepareDisplay				; Copy Expanded storage to Display format
			movf	SEQDURATION,W				; Set-up frame duration
			call	DisplayFrame				; Display frame
			incf	CURFRAME,F					; Increment for next frame
			goto	FrameLoop

FrameLoopEnd
			bcf		PIXELFLAGS,FLAG_PIXEL_DONE	; Reset these
			bcf		PIXELFLAGS,FLAG_FIRST_PASS	; Now on at least second pass of animation
			btfss	PIXELFLAGS,FLAG_WRAP_STYLE	; Reset 'pixels off' flag for single pass style
			bcf		PIXELFLAGS,FLAG_PIXELS_OFF
			decfsz	SEQREPEAT,F
			goto	RepeatLoop
			
			incf	CURSEQ,F
			goto	ListLoop
			
			
			
		;---------------------------------------------
		;	PrepareDisplay
		;	Copied the data from EXPANDED storage into
		;	the DISPLAY storage, making any adjustments
		;	to the data layout as necessary.
		;	
		;---------------------------------------------
PrepareDisplay
			movf	EXPANDED,w
			movwf	DISPLAY

			movf	EXPANDED+1,w
			movwf	DISPLAY+1

			movf	EXPANDED+2,w
			movwf	DISPLAY+2

			movf	EXPANDED+3,w
			movwf	DISPLAY+3

			movf	EXPANDED+4,w
			movwf	DISPLAY+4

#if MIRROR_HORIZONTAL
			movlw	DISPLAY
			call	CopyToWorkBuffer
			call	MirrorHorzWorkBuffer
			movlw	DISPLAY
			call	CopyFromWorkBuffer
#endif

#if (PIXELDATA_DIM==COL_VALUE)
			movlw	DISPLAY
			call	CopyToWorkBuffer
			call	RotateWorkBuffer
			movlw	DISPLAY
			call	CopyFromWorkBuffer
#endif

			; Align bits to port, if required
#if PIXELDATA_NEEDS_ALIGN_TO_PORT
			movf	DISPLAY,W
			call	AlignToPort
			movwf	DISPLAY

			movf	DISPLAY+1,W
			call	AlignToPort
			movwf	DISPLAY+1

			movf	DISPLAY+2,W
			call	AlignToPort
			movwf	DISPLAY+2

			movf	DISPLAY+3,W
			call	AlignToPort
			movwf	DISPLAY+3

			movf	DISPLAY+4,W
			call	AlignToPort
			movwf	DISPLAY+4

#endif


			; Invert the bits for active high pixel data
#if !PIXELDATA_ACTIVE_HIGH
			movlw	0x4f
			xorwf	DISPLAY,F
			xorwf	DISPLAY+1,F
			xorwf	DISPLAY+2,F
			xorwf	DISPLAY+3,F
			xorwf	DISPLAY+4,F
#else
			nop
			nop
			nop
			nop
			nop
			nop
#endif

			return

		;---------------------------------------------
		;	DisplayFrame
		;	Display the current frame in expanded storage
		;	for 'w' iterations multiplied by 0x10.
		;	The 0x10 factor was added to slow down everything
		;---------------------------------------------
DisplayFrame
			movwf	FRAMECOUNT
			
LoopDFA		movlw	0x02 		; 2008.09.20: Changed from 0x08 to 0x02
			movwf	FRAMEBCOUNT
LoopDFB		call	DisplaySingleFrame
			decfsz	FRAMEBCOUNT,F
			goto	LoopDFB
			decfsz	FRAMECOUNT,F
			goto	LoopDFA
			return
		
		;---------------------------------------------
		;	DisplaySingleFrame
		;	Display info in EXPANDED storage
		;	
		;
		;---------------------------------------------
DisplaySingleFrame
			call	DisplayOff

			movf	DISPLAY,W		; Show row-or-col #1 (index 0)
			movwf	PIXELDATA_PORT
			movlw	MULTIPLEX_ZERO
			movwf	MULTIPLEX_PORT
			call	MultiplexDelay
			
			call	DisplayOff

			movf	DISPLAY+1,W		; Show row-or-col #2 (index 1)
			movwf	PIXELDATA_PORT
			movlw	MULTIPLEX_ONE
			movwf	MULTIPLEX_PORT
			call	MultiplexDelay
			
			call	DisplayOff

			movf	DISPLAY+2,W		; Show row-or-col #3 (index 2)
			movwf	PIXELDATA_PORT
			movlw	MULTIPLEX_TWO
			movwf	MULTIPLEX_PORT
			call	MultiplexDelay
			
			call	DisplayOff

			movf	DISPLAY+3,W		; Show row-or-col #4 (index 3)
			movwf	PIXELDATA_PORT
			movlw	MULTIPLEX_THREE
			movwf	MULTIPLEX_PORT
			call	MultiplexDelay

			call	DisplayOff

			movf	DISPLAY+4,W		; Show row-or-col #5 (index 4)
			movwf	PIXELDATA_PORT
			movlw	MULTIPLEX_FOUR
			movwf	MULTIPLEX_PORT
			call	MultiplexDelay

			return
			
		;---------------------------------------------
		;	DisplayOff
		;	Turn-off all display lines
		;	
		;
		;---------------------------------------------
DisplayOff
			movlw	0xff
#if MULTIPLEX_ACTIVE_HIGH
			clrf	MULTIPLEX_PORT
#else
			movwf	MULTIPLEX_PORT
#endif
#if PIXELDATA_ACTIVE_HIGH
			clrf	PIXELDATA_PORT
#else
			movwf	PIXELDATA_PORT
#endif
			return

		;---------------------------------------------
		;	AlignToPort
		;	Moves bit 4 to bit 6 in w
		;	'xxxEDCBA' --> 'xExxDCBA'
		;---------------------------------------------
AlignToPort
			andlw	b'00011111'	; Clear don't care bits
			movwf	TEMP		; ..and store
			andlw	b'00010000' ; Leave only bit [4:4]
			movwf	TEMP2		; ..and store
			bcf		STATUS,C	; Clear carry
			rlf		TEMP2,F		; ..and rotate twice
			rlf		TEMP2,F
			movlw	b'00001111'	; Leave only bits [3:0]
			andwf	TEMP,W		; ..from original in W
			iorwf	TEMP2,W		; Merge with bit [6:6]
			return

		;---------------------------------------------
		;	ExpandPacked
		;	Expands data in PACKED storage and
		;	stores it in EXPANDED storage
		;
		;	0 oooAAAAA	->	0 ...AAAAA
		;	1 -#####oo		1 ...ooooo
		;	2 ^^^^----		2 ...#####
		;	3 .......^		3 ...-----
		;					4 ...^^^^^
		;
		;---------------------------------------------
ExpandPacked
			bcf		STATUS,C	; Keep this clear

			; --0-- 
			;	"...43210"  >> "...43210"
									; Bits "04 03 02 01 00"
			movf	PACKED,W		; Take five lowest bits   "...XXXXX"
			andlw	0x1f			; ..from PACKED and store
			movwf	EXPANDED		; ..in EXPANDED
			
			; --1--
			; "210....." >>   "...43210"
			; "......43" 
									; Bits "09 08 07 06 05"
			swapf	PACKED,W		; SWAP three high bits in "XXX....."
			andlw	0x0e			; ..PACKED, mask, store
			movwf	EXPANDED+1		; ..store
			rrf		EXPANDED+1,f	; ..rotate done one
			swapf	PACKED+1,W		; "......XX" -> "..XX...."
			andlw	0x30
			movwf	TEMP
			rrf		TEMP,W
			iorwf	EXPANDED+1,W
			movwf	EXPANDED+1
			
			; --2--
			; ".43210.." >>   "...43210"
			
			movf	PACKED+1,W
			andlw	0x7c
			movwf	EXPANDED+2
			rrf		EXPANDED+2,F
			rrf		EXPANDED+2,W
			movwf	EXPANDED+2
			
			; --3--
			; "0......." >>   "...43210"
			; "....4321" 
									; Bits "09 08 07 06 05"
			movf	PACKED+2,W		; Takes bits "4321"
			andlw	0x0f			; ..and place in
			movwf	EXPANDED+3		; ..destination and
			rlf		EXPANDED+3,F	; ..rotate into place
			btfsc	PACKED+1,7		; ..check bit "0" and
			bsf		EXPANDED+3,0	; ..set if necessary
			movf	EXPANDED+3,W
			movwf	EXPANDED+3
			
			; --4--
			; "3210...." >>   "...43210"
			; ".......4" 
									; Bits "09 08 07 06 05"
			swapf	PACKED+2,W		; Takes bits "3210"
			andlw	0x0f			; ..and place in
			movwf	EXPANDED+4		; ..destination
			btfsc	PACKED+3,0		; ..check bit "4" and
			bsf		EXPANDED+4,4	; ..set if necessary
			movf	EXPANDED+4,W
			movwf	EXPANDED+4
			return
						
		;---------------------------------------------
		;	Multiplex Delay
		;	Pre-set delay to be called between row
		;	updates on display
		;---------------------------------------------
MultiplexDelay	
			movlw	MULTIPLEX_DELAY
			goto	SmallDelay

		;---------------------------------------------
		;	Small Delay
		;	Max delay @ 4Mhz: 1.275 msec
		;---------------------------------------------
SmallDelay	
			movwf	COUNTER
LoopSD		nop
			nop
			decfsz	COUNTER,F
			goto	LoopSD
			return
		;---------------------------------------------
		;	LoadFrameData
		;	Called by frame-by-frame animations
		;	'Routine index' passed in w
		;---------------------------------------------
LoadFrameData
			bsf		SEQFLAGS,FLAG_GET_DATA
			bcf		STATUS,C				; Clear carry so we can use
			movf	CURFRAME,W				; ..RotateLeft twice to
			movwf	DATAINDEX				; ..multiple by four
			rlf		DATAINDEX,F
			rlf		DATAINDEX,F
			call	AnimationDispatch		; Get first data byte
			movwf	PACKED					; ..and stored in packed array
			btfsc	SEQFLAGS,FLAG_DATA_DONE	; Did sequence end?
			goto	EndLFD					; ..then leave
			incf	DATAINDEX,F				; Increment index to
			call	AnimationDispatch		; ..get second byte
			movwf	PACKED+1				; ..and store		
			incf	DATAINDEX,F				; Increment index to
			call	AnimationDispatch		; ..get third byte
			movwf	PACKED+2				; ..and store		
			incf	DATAINDEX,F				; Increment index to
			call	AnimationDispatch		; ..get fourth byte
			movwf	PACKED+3				; ..and store		
EndLFD		bcf		SEQFLAGS,FLAG_GET_DATA
			return


		;---------------------------------------------
		;	LoadPixelListData 
		;---------------------------------------------
LoadPixelListData
			bsf		SEQFLAGS,FLAG_GET_DATA		; This flags tells the animation routines that we
												; ..are going to ask for data
			btfsc	PIXELFLAGS,FLAG_PIXEL_DONE	; Did we already run out of pixel data?
			goto	TestRmvLPLD					; ..then just remove pixels
			movf	CURFRAME,W					; Get current frame then 
			call	GetPixelValue				; ..use that to get pixel
			btfss	SEQFLAGS,FLAG_DATA_DONE		; Did sequence end?
			goto	ContLPLD					; ..if not then continue setting pixels
			btfss	PIXELFLAGS,FLAG_WRAP_STYLE	; If we are not wrap style
			goto	DataEndLPLD					; ..then jump ahead to adjust flags and remove pixels
			decfsz	SEQREPEAT,W					; For wrap style if this is last repeat then
												; ..it is treated a 'single pass' and we keep going by
												; ..resetting flags.
			goto	EndLPLD						; Otherwise we end the sequence now and start over
DataEndLPLD	bcf		SEQFLAGS,FLAG_DATA_DONE		; Keep this reset at this point
			bsf		PIXELFLAGS,FLAG_PIXEL_DONE	; Note that pixel data is done
			goto	TestRmvLPLD					; ..and only worry about removing pixels
ContLPLD	call	MakeMaskFromPixel			; Create BITMASK
			call	UpdateFSRFromPixel			; Calculate value for indexed addressing
			movf	BITMASK,W					; Get mask
			iorwf	INDF,F						; Update packed memory using indexed addr.

TestRmvLPLD	btfsc	PIXELFLAGS,FLAG_PIXELS_OFF	; End was already found, so skip test
			goto	RemoveLPLD					; ..and continue removing pixels
			
			movf	CURFRAME,W					; Determine if runlength exceeded
			movwf	TEMP						; ..by moving current frame to temp
			movf	RUNLENGTH,W					; ..and subtracting runlength
			subwf	TEMP,W						;
			btfss	STATUS,C					; If carry is set, then runlength
			goto	EndLPLD						; ..has not been reached, so return
			bsf		PIXELFLAGS,FLAG_PIXELS_OFF	; Set flag to indicate that we are now removing pixels
			clrf	PIXELOFF
			
RemoveLPLD	movf	PIXELOFF,W					; Use index in PIXELOFF
			call	GetPixelValue				; ..to get pixel to turn-off
			btfss	SEQFLAGS,FLAG_DATA_DONE		; Did sequence end? Then all pixels should
			goto	RmvSkipLDLD					; If not, then continue removing pixels.
			clrf	PIXELOFF					; Otherwise, reset PIXELOFF index
			btfss	PIXELFLAGS,FLAG_WRAP_STYLE	; If we are not doing wrap style,
			goto	EndLPLD						; ..then we can return
			btfss	PIXELFLAGS,FLAG_PIXEL_DONE	; Otherwise, for wrap style, if we are still
			goto	RmvEndLPLD					; ..setting pixels, then we continue
			goto	EndLPLD						; ..otherwise we are doine
RmvEndLPLD	bcf		SEQFLAGS,FLAG_DATA_DONE		; Otherwise reset the flag
			goto	RemoveLPLD					; ..and remove pixel at zero
RmvSkipLDLD	call	MakeMaskFromPixel			; The bitmask
			call	UpdateFSRFromPixel			; Calculate value for indexed addressing
			movlw	0xff						; Get inverted mask
			xorwf	BITMASK,W					; 
			andwf	INDF,F						; Update packed memory using indexed addr.
			incf	PIXELOFF,F					; Increment PIXELOFF index
			
EndLPLD		bcf		SEQFLAGS,FLAG_GET_DATA

			return


		;---------------------------------------------
		;	LoadStripFrame
		;
		;	ROW_INDEX: represents row of frame data to place at
		;	row zero of the display frame.
		;
		;	TMP_INDEX: Index for current row of frame. Starts with ROW_INDEX
		;	value and incremented for each row of frame
		;
		;	ROW_COUNT: Counter for the rows of the frame
		;
		;	FLAG_DATA_DONE: Sets this flag if ROW_INDEX points to first 
		;	row past row data. Note that this means that this routine should
		;	clear this flag if it is set for frame rows 2nd thru last.
		;---------------------------------------------
LoadStripFrame
			; If ROW_INDEX is less than zero we are at or near the beginning
			; ..and we know that we have not reach the end. Also, we test this
			; ..here so we don't pass a negative number to the AnimationDispatch
			; ..routine a few lines later
			btfsc	ROW_INDEX,7					; High bit set means negative
			goto	LSF_LpInit
			; If ROW_INDEX points to the end of the frame data, then we are done
			; ..and we leave this flag set and we exit
			; Note: By testing for ROW_INDEX to point at the end of frame data, we are
			; ..going to stop when there is the last row of data still in the frame.
			; ..If we want to stop *after* the frame is clear, we need to test with ROW_INDEX+1
			; ..instead of ROW_INDEX
			movf	ROW_INDEX,w
			movwf	DATAINDEX				; Store w
			call	AnimationDispatch
			btfsc	SEQFLAGS,FLAG_DATA_DONE
			goto	LSF_Exit
LSF_LpInit
			; Set-up loop
			movf	ROW_INDEX,w					; Copy ROW_INDEX to TMP_INDEX
			movwf	TMP_INDEX
			movlw	FRAME_SIZE					; Init ROW_COUNT
			movwf	ROW_COUNT
			movlw	EXPANDED					; Indirect addressing for destination data
			movwf	FSR
			; Loop
LSF_Loop
			clrf	INDF						; Always clear the row
			btfsc	TMP_INDEX,7					; If index is negative, then leave row blank
			goto	LSF_LpBottom
			btfsc	PIXELFLAGS,FLAG_PIXEL_DONE	; If data already ended, then leave row blank
			goto	LSF_LpBottom
			movf	TMP_INDEX,w					; Set DATAINDEX with 
			movwf	DATAINDEX					; ..the TMP_INDEX and call to get the data
			call	AnimationDispatch
			btfss	SEQFLAGS,FLAG_DATA_DONE		; Did we run out of data?
			goto	LSF_LpStore					; ..no, so goto 'store' code
			; Handle end-of-data
			btfss	PIXELFLAGS,FLAG_WRAP_STYLE	; If 'single pass' style, then 
			goto	LSF_LpDataEnd				; ..goto to code that handles that
			decfsz	SEQREPEAT,W					; For 'wrap style' if this is last repeat then
			goto	$+2							; ..it is treated a 'single pass' and we keep going by
			goto	LSF_LpDataEnd				; ..jumping to the 'data end' code
			clrf	TMP_INDEX					; For 'wrap style', clear the TMP_INDEX and the
			clrf	DATAINDEX					; ..DATAINDEX and ask for the first row of data
			call	AnimationDispatch
			bcf		SEQFLAGS,FLAG_DATA_DONE		; Clear 'data done' flag so top loop will not terminate
			goto	LSF_LpStore
LSF_LpDataEnd
			bsf		PIXELFLAGS,FLAG_PIXEL_DONE	; Indicate that we have run out of pixel data
			bcf		SEQFLAGS,FLAG_DATA_DONE		; Clear 'data done' flag so top loop will not terminate
			goto	LSF_LpBottom
LSF_LpStore
			movwf	INDF						; Store result in expanded store
LSF_LpBottom
			incf	FSR,f						; Increment destination
			incf	TMP_INDEX,f					; Increment row data index
			decfsz	ROW_COUNT,f
			goto LSF_Loop
			bcf		PIXELFLAGS,FLAG_PIXEL_DONE	; Clear this at the end of every frame.

LSF_Exit
			return

		;=============================================
		;	Page 0x100
		;=============================================
		;	org 0x100

		;---------------------------------------------
		;	LoadStripData 
		;---------------------------------------------
LoadStripData
			bsf		SEQFLAGS,FLAG_DATA_EXPANDED	; Set this flag to indicate that we are updating
												; ..the 'expanded' data and not the 'packed' data
			bsf		SEQFLAGS,FLAG_GET_DATA		; This flags tells the animation routines that we
												; ..are going to ask for data
			movlw	-FRAME_SIZE					; Convert 'current frame' to 'row index' where..
			btfss	PIXELFLAGS,FLAG_WRAP_STYLE	; ..'index' is an index into data. When 'index' is less
			goto	LSF_AddIndex				; ..than zero, then data to load will be zero.
			btfss	PIXELFLAGS,FLAG_FIRST_PASS	
			movlw	0

LSF_AddIndex
			addwf	CURFRAME,W
			movwf	ROW_INDEX
			call	LoadStripFrame				; Load data. This routine will set FLAG_DATA_DONE
												; ..if ROW_INDEX points to first row past row data
			btfsc	PIXELFLAGS,FLAG_HORIZONTAL	; If the strip is horizontal, then rotate
			call	RotateExpanded				; ..the expanded storage
			bcf		SEQFLAGS,FLAG_GET_DATA		; No longer asking for data
			return

		;---------------------------------------------
		;	CopyToWorkBuffer
		;
		;	W contains address of buffer to copy
		;---------------------------------------------
CopyToWorkBuffer
			movwf	FSR

			movf	INDF,W
			movwf	WORK
			incf	FSR,f

			movf	INDF,W
			movwf	WORK+1
			incf	FSR,f

			movf	INDF,W
			movwf	WORK+2
			incf	FSR,f

			movf	INDF,W
			movwf	WORK+3
			incf	FSR,f

			movf	INDF,W
			movwf	WORK+4

			return

		;---------------------------------------------
		;	CopyFromWorkBuffer
		;
		;	W contains address of buffer to copy
		;---------------------------------------------
CopyFromWorkBuffer
			movwf	FSR

			movf	WORK,W
			movwf	INDF
			incf	FSR,f

			movf	WORK+1,W
			movwf	INDF
			incf	FSR,f

			movf	WORK+2,W
			movwf	INDF
			incf	FSR,f

			movf	WORK+3,W
			movwf	INDF
			incf	FSR,f

			movf	WORK+4,W
			movwf	INDF

			return

		;---------------------------------------------
		;	MirrorHorzWorkBuffer
		;---------------------------------------------
MirrorHorzWorkBuffer
			movlw	FRAME_SIZE
			movwf	TEMP

			movlw	WORK
			movwf	FSR

MHWB_RotateLoop
			clrf	TEMP2

			rrf		INDF,f			; Shift low bit off of each row
			rlf		TEMP2,f			; ..and shift onto new row

			rrf		INDF,f
			rlf		TEMP2,f

			rrf		INDF,f
			rlf		TEMP2,f

			rrf		INDF,f
			rlf		TEMP2,f

			rrf		INDF,f
			rlf		TEMP2,f

			movf	TEMP2,w
			movwf	INDF

			incf	FSR,f

			decfsz	TEMP,f
			goto	MHWB_RotateLoop


			return

		;---------------------------------------------
		;	RotateWorkBuffer
		;---------------------------------------------
RotateWorkBuffer
			movlw	FRAME_SIZE
			movwf	TEMP

			movlw	ROTATED
			movwf	FSR

RWB_RotateLoop
			clrf	TEMP2

			rrf		WORK,f		; Shift low bit off of each row
			rlf		TEMP2,f			; ..and shift onto new row

			rrf		WORK+1,f
			rlf		TEMP2,f

			rrf		WORK+2,f
			rlf		TEMP2,f

			rrf		WORK+3,f
			rlf		TEMP2,f

			rrf		WORK+4,f
			rlf		TEMP2,f

			movf	TEMP2,w
			movwf	INDF
			incf	FSR,f

			decfsz	TEMP,f
			goto	RWB_RotateLoop


			movf	ROTATED,w		; Move rotated values to
			movwf	WORK		; ..display location
			movf	ROTATED+1,w
			movwf	WORK+1
			movf	ROTATED+2,w
			movwf	WORK+2
			movf	ROTATED+3,w
			movwf	WORK+3
			movf	ROTATED+4,w
			movwf	WORK+4
			return


		;---------------------------------------------
		;	RotateExpanded
		;---------------------------------------------
RotateExpanded
			movlw	FRAME_SIZE
			movwf	TEMP

			movlw	ROTATED
			movwf	FSR

RE_RotateLoop
			clrf	TEMP2

			rrf		EXPANDED,f		; Shift low bit off of each row
			rlf		TEMP2,f			; ..and shift onto new row

			rrf		EXPANDED+1,f
			rlf		TEMP2,f

			rrf		EXPANDED+2,f
			rlf		TEMP2,f

			rrf		EXPANDED+3,f
			rlf		TEMP2,f

			rrf		EXPANDED+4,f
			rlf		TEMP2,f

			movf	TEMP2,w
			movwf	INDF
			incf	FSR,f

			decfsz	TEMP,f
			goto	RE_RotateLoop


			movf	ROTATED,w		; Move rotated values to
			movwf	EXPANDED		; ..display location
			movf	ROTATED+1,w
			movwf	EXPANDED+1
			movf	ROTATED+2,w
			movwf	EXPANDED+2
			movf	ROTATED+3,w
			movwf	EXPANDED+3
			movf	ROTATED+4,w
			movwf	EXPANDED+4
			return

		;---------------------------------------------
		;	MakeMaskFromPixel 
		;
		;	Uses value in PIXEL to create a mask in
		;	BITMASK where the Nth bit is set and
		;	all other bits are zero.
		;---------------------------------------------
MakeMaskFromPixel
			movlw	1						; Set-up bitmask
			movwf	BITMASK
			bcf		STATUS,C				; Clear carry for later rotate
			movlw	0x7						; Get low three bits
			andwf	PIXEL,W					; ..into w
			movwf	TEMP					; ..and move to TEMP
			btfsc	STATUS,Z				; If zero, then mask already set
			goto	SkipMMFP				; ..so skip rotate loop
LoopMMFP	rlf		BITMASK,F				; Rotate mask 
			decfsz	TEMP,F					; ..until number of bits satisfied
			goto	LoopMMFP
SkipMMFP	nop
			return		
			

		;---------------------------------------------
		;	UpdateFSRFromPixel
		;
		;	Uses value in PIXEL to 
		;	
		;	
		;---------------------------------------------
UpdateFSRFromPixel 
			movf	PIXEL,W						; Set-up indirect addressing
			andlw	0x18
			movwf	FSR
			bcf		STATUS,C
			rrf		FSR,F
			rrf		FSR,F
			rrf		FSR,F
			movlw	PACKED
			addwf	FSR,F
			bcf		STATUS,IRP					; (Clear 9th bit of indirect addressing Status:<7>)
			return

		;---------------------------------------------
		;	GetPixelValue 
		;
		;	Use the value in W as an index in the 
		;	pixel list animation data. Get the Wth
		;	item from list.
		;
		;	Populates DATAINDEX with W 
		;---------------------------------------------
GetPixelValue
			movwf	DATAINDEX				; Store w
			call	AnimationDispatch		; Get data. Uses DATAINDEX
			movwf	PIXEL					; Store value
			return




